/*
 * Dictionary.java
 *
 * Created on 14  2007 ., 0:29
 *
 */
import java.io.*;
//import java.util.Hashtable;
import java.util.Vector;
/**
 *
 * @author V*D*V
 */
public class Dictionary// extends Hashtable
{
    private static final String DICTIONARY_NAME = "d.txt";
    private static final String WORD_SEPARATOR = "\t";
    //private static final String ACCENT_SYMBOL = "^";
    private int m_enc = Reader.UNICODE;
    //
    private Vector m_vTable = new Vector();
    private static final int WORD_EQUAL = 0;
    private static final int WORD_PART = 1;
    private static final int WORD_START = 2;
    private static final int WORD_END = 3;
    //
    private static final char[] m_chDeaf = {'', '', '', '', '', '', '', ''};
    private static final char[] m_chSonant = {'', '', '', '', '', ''};
    private static final char[] m_chVowel = {'', '', '', '', '', '', '', '', '', ''};

    private class Equal
    {
        public String source;
        public String target;
        public int part;
        public Equal(String src, String trg, int p)
        {
            source = src;
            target = trg;
            part = p;
        }
    }

    public Dictionary()
    {
        super();
    }
    public void loadDictionary()
    {
        InputStream is = getClass().getResourceAsStream( DICTIONARY_NAME );
        if ( is == null ) return;
        try
        {
            int d = is.read();
            m_enc = Reader.UNICODE;
            if ( d < 0xff )
            {
                m_enc = Reader.ANSI;
                is.close();
                is = getClass().getResourceAsStream( DICTIONARY_NAME );
            }
            else
            {
                d = is.read();
            }
            while ( true )
            {
                String row = getNextRow( is );
                if ( row == null ) return;
                int separator = row.indexOf( WORD_SEPARATOR );
                if ( separator < 1 ) continue;
                String word = row.substring( 0, separator );
                String voc = row.substring( separator + 1 );
                word = word.trim();
                voc = voc.trim();
                word = word.toLowerCase();
                //if ( word.length() > 0 && voc.length() > 0 ) { put( word, voc ); }
                separator = WORD_EQUAL;
                if ( word.endsWith( "*" ) )
                {
                    word = word.substring( 0, word.length() - 1 );
                    separator = WORD_START;
                }
                if ( word.startsWith( "*" ) )
                {
                    if ( separator == WORD_START ) separator = WORD_PART;
                    else separator = WORD_END;
                    word = word.substring( 1, word.length() );
                }
                Equal eq = new Equal( word, voc, separator );
                m_vTable.addElement( eq );
            }
        }
        catch (IOException ie){}
    }
    private String getNextRow(InputStream is) throws IOException, UnsupportedEncodingException
    {
        String str = "";
        do
        {
            char ch = ' ';
            if ( m_enc == Reader.UNICODE )
            {
                byte[] data = new byte[2];
                if ( is.read( data ) <= 0 )
                {
                    if ( str.equals( "") ) return null;
                    return str;
                }
                ch = (char)( ( data[0] & 0xff ) | ( ( data[1] << 8 ) & 0xff00 ) );
            } else if ( m_enc == Reader.ANSI )
            {
                int data = is.read( );
                if ( data <= 0 )
                {
                    if ( str.equals( "") ) return null;
                    return str;
                }
                if ( data < 128) ch = (char)data;
                else ch = Reader.WIN1251_TO_UNICODE.charAt( data - 128 );
            } else
            {
                int data = is.read( );
                if ( data <= 0 ) return null;
            }
            if ( ch == '\r' || ch == '\n' ) break;
            str += String.valueOf( ch );
        }
        while ( true );
        return str;
    }
    public String getPronunciation(String word)
    {
        String original = "";
        String end = "";
        for ( int i = 0; i < word.length(); i++ )
        {
            char ch = word.charAt( i );
            if ( ( ch >= '' && ch <= '' ) || ch == '' )
            {
                ch += '' - '';
            }
            if ( ( ch < '' || ch > '' ) && ch != '' )
            {
                end = word.substring( i );
                //original = original.substring( 0, i + 1 );
                break;
            }
            original += String.valueOf( ch );
        }
        original = original.toLowerCase();
        //
        //String pron = (String)get( original );
        String pron = replaceFromDictionary( original );
        //
        if ( pron != null ) { return pron + end; }
        if ( original.length() < 2 ) { return word.toLowerCase(); }
        //apply pronunciation rules
        int i = original.indexOf( "" );
        if ( i < 0 ) i = original.indexOf( "" );
        if ( i < 0 ) i = original.indexOf( "" );
        if ( i < 0 ) i = original.indexOf( "" );
        if ( i < 0 ) i = original.indexOf( "" );
        if ( i < 0 ) i = original.indexOf( "" );
        if ( i >= 0 ) original = original.substring( 0, i + 1 ) + original.substring( i + 2 );
        //
        i = original.indexOf( "" );
        if ( i < 0 ) i = original.indexOf( "" );
        if ( i >= 0 ) original = original.substring( 0, i ) + original.substring( i + 1 );
        //
        //if ( original.length() > 2 )
        {
            for ( i = 0; i < m_chVowel.length; i++ )
            {
                if ( original.charAt( original.length() - 1 ) == m_chVowel[i] ) break;
            }
            if ( i >= m_chVowel.length
              && original.charAt( original.length() - 1 ) == original.charAt( original.length() - 2 ) )
            { original = original.substring( 0, original.length() - 1 ); }
        }
        //
        for ( i = 0; i < original.length() - 1; i++ )
        {
            for ( int s = 0; s < m_chSonant.length; s++ )
            {
                if ( original.charAt( i ) == m_chSonant[s] )
                {
                    for ( int d = 0; d < m_chDeaf.length; d++ )
                    {
                        if ( original.charAt( i + 1 ) == m_chDeaf[d] )
                        {
                            original = original.substring( 0, i ) + String.valueOf( m_chDeaf[s] ) + original.substring( i + 1 );
                        }
                    }
                }
            }
        }
        //
        for ( int s = 0; s < m_chSonant.length; s++ )
        {
            if ( original.charAt( original.length() - 1 ) == m_chSonant[s] )
            {
                original = original.substring( 0, original.length() - 1 ) + String.valueOf( m_chDeaf[s] );
            }
        }
        //
        return original + end;
    }

    private String replaceFromDictionary(String word)
    {
        for ( int i = 0; i < m_vTable.size(); i++ )
        {
            String src = ((Equal)m_vTable.elementAt( i ) ).source;
            if ( ((Equal)m_vTable.elementAt( i ) ).part == WORD_PART )
            {
                int indx = word.indexOf( src );
                if ( indx >= 0 )
                {
                    String end = "";
                    if ( indx + src.length() < word.length() ) end = word.substring( indx + src.length() );
                    word = word.substring( 0, indx ) + ((Equal)m_vTable.elementAt( i ) ).target + end;
                    return word;
                }
            } else if ( ((Equal)m_vTable.elementAt( i ) ).part == WORD_EQUAL )
            {
                if ( word.equals( src ) )
                {
                    return ((Equal)m_vTable.elementAt( i ) ).target;
                }
            } else if ( ((Equal)m_vTable.elementAt( i ) ).part == WORD_START )
            {
                if ( word.startsWith( src ) )
                {
                    word = ((Equal)m_vTable.elementAt( i ) ).target + word.substring( src.length() );
                    return word;
                }
            } else if ( ((Equal)m_vTable.elementAt( i ) ).part == WORD_END )
            {
                if ( word.endsWith( src ) )
                {
                    word = word.substring( 0, word.length() - src.length() ) + ((Equal)m_vTable.elementAt( i ) ).target;
                    return word;
                }
            }
        }
        return null;
    }
}
