import java.io.*;
import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.GameCanvas;
import javax.microedition.media.*;
import javax.microedition.media.control.VolumeControl;
import javax.microedition.midlet.MIDlet;
import javax.microedition.rms.*;
/*
 * Screen.java
 *
 * Created on 12  2007 ., 0:01
 *
 */

/**
 *
 * @author V*D*V
 */
public class Screen extends GameCanvas implements CommandListener, PlayerListener, Runnable
{
    private static final int PROGRESS_BAR_WIDTH = 2;
    private static final int TEXT_LEFT_SHIFT = 1;
    private Command cmdStart = new Command( "Play", Command.SCREEN, 1 );
    private Command cmdStop = new Command( "Stop", Command.SCREEN, 2 );
    private Command cmdAutoScroll = new Command( "AutoScroll", Command.SCREEN, 3 );
    private Command cmdNextRow = new Command( "NextRow", Command.SCREEN, 4 );
    private Command cmdNextScreen = new Command( "NextScreen", Command.SCREEN, 5 );
    private Command cmdAbout = new Command( "About", Command.SCREEN, 6 );
    private Command cmdExit = new Command( "Exit", Command.SCREEN, 7 );
    private MIDlet m_midlet;
    private Display m_display;
    private boolean m_bExit = false;
    private Thread m_thread;
    private Player m_player = null;
    private Object m_waitRead = new Object();
    private boolean m_bNextRead = false;
    private Object m_waitScroll = new Object();
    private long m_lScrollWaitTime = 0;
    private boolean m_bVoice = false;
    private Graphics m_g;
    private Font m_font;
    private InputStream m_is = null;
    private int m_encoding = Reader.UNICODE;
    private int m_iBookLength = 0;
    private int m_iShowPosition = 0;
    private int m_iNextRowPosition = 0;
    private boolean m_bEndBook = false;
    //private int m_iReadPosition = 0;
    private int m_iMaxRows = 0;
    //private int m_iLettersOnScreen = 0;
    private int m_iTextWidth = 0;
    private String[] m_pStrRows;
    private int[] m_pPositionRows;
    private String m_strNextWord;
    private Word m_word;
    private Dictionary m_dic;
/**
 *internal class
 */
    private class Suggestion
    {
        public String m_str;// suggestion
        public int m_y;// y position of end of suggestion
        public int m_x;// x position of end of suggestion in the string
    }
    public Screen( MIDlet m, Display d )
    {
        super( false );
        m_midlet = m;
        m_display = d;
        addCommand( cmdStart );
        addCommand( cmdStop );
        addCommand( cmdAutoScroll );
        addCommand( cmdNextRow );
        addCommand( cmdNextScreen );
        addCommand( cmdAbout );
        addCommand( cmdExit );
        setCommandListener( this );
        m_thread = new Thread( this );
        m_thread.start();
    }

    public void keyPressed(int keyCode)
    {
        switch( keyCode )
        {
            case KEY_NUM1://start
                cmdStart();
                break;
            case KEY_NUM2://stop
                cmdStop();
                break;
            case KEY_NUM3://auto scroll
                cmdAutoScroll();
                break;
            case KEY_NUM4://next row
                cmdNextRow();
                break;
            case KEY_NUM5://next screen
                cmdNextScreen();
                break;
            case KEY_NUM6://
                break;
            case KEY_NUM7://
                break;
            case KEY_NUM8://
                break;
            case KEY_NUM9://
                break;
            case KEY_NUM0://exit
                cmdExit();
                break;
        }
    }
    public void commandAction(Command c, Displayable displayable)
    {
        if( c == cmdStart ) { cmdStart(); }
        else if( c == cmdStop ) { cmdStop(); }
        else if( c == cmdAutoScroll ) { cmdAutoScroll(); }
        else if( c == cmdNextRow ) { cmdNextRow(); }
        else if( c == cmdNextScreen ) { cmdNextScreen(); }
        else if( c == cmdAbout ) { cmdAbout(); }
        else if( c == cmdExit ) { cmdExit(); }
    }
    private void cmdStart()
    {
        synchronized( m_waitScroll ) { m_waitScroll.notify(); }
        synchronized( m_waitRead ) { m_waitRead.notify(); }
        m_lScrollWaitTime = 1;
        m_bVoice = true;
    }
    private void cmdStop()
    {
        stopPlayer();
        m_lScrollWaitTime = 0;
        m_bVoice = false;
    }
    private void cmdAutoScroll()
    {
        m_lScrollWaitTime = 50 * m_iTextWidth / m_font.charWidth( ' ' );//letters * 0.05 sec
        m_bVoice = false;
        synchronized( m_waitRead ) { m_waitRead.notify(); }
        synchronized( m_waitScroll ) { m_waitScroll.notify(); }
    }
    private void cmdNextRow()
    {
        stopPlayer();
        m_lScrollWaitTime = 0;
        nextRow();
    }
    private void cmdNextScreen()
    {
        stopPlayer();
        m_lScrollWaitTime = 0;
        nextScreen();
    }
    private void cmdAbout()
    {
        Alert msg = new Alert( "About Reader" );
        msg.setTimeout( Alert.FOREVER );
        msg.setString( "Control:\n" +
                       "Play: 1\n" +
                       "Stop: 2\n" +
                       "Auto Scroll: 3\n" +
                       "Next Row: 4\n" +
                       "Next Screen: 5\n\n" +
                       "A lot of thanks:\n" +
                       "Voice: Evgeniya Shestakova\n" +
                       "Record by: Semen Oblomkin\n\n" +
                       "Vendor: (c) 2007 V*D*V\n" +
                       "Version: 1.0\n" +
                       "Support: gamemobile@yahoo.com" );
        m_display.setCurrent( msg, m_display.getCurrent() );
    }
    public void cmdExit()
    {
        stopPlayer();
        m_bExit = true;
        synchronized( m_waitRead ) { m_waitRead.notify(); }
        m_lScrollWaitTime = 1;
        synchronized( m_waitScroll ) { m_waitScroll.notify(); }
        m_midlet.notifyDestroyed();
    }
    public void paint(Graphics g)
    {
        flushGraphics();
    }
    public void playerUpdate(Player player, String event, Object data)
    {
        if ( event.equals( PlayerListener.END_OF_MEDIA ) )
        {
            stopPlayer();
            synchronized( m_waitRead ) { m_waitRead.notify(); }
            m_bNextRead = true;
        } else if ( event.equals( PlayerListener.STARTED ) )
        {
        } else if ( event.equals( PlayerListener.STOPPED ) )
        {
        } else if ( event.equals( PlayerListener.VOLUME_CHANGED ) )
        {
        }
    }

    public void run()
    {
        init();
        m_g.drawString( "prepare text", getWidth() / 2, getHeight() / 2 + m_font.getHeight(), m_g.HCENTER | m_g.BOTTOM );
        repaint();
        prepareText();
        drawText();
        try { synchronized( m_waitScroll ) { m_waitScroll.wait( m_lScrollWaitTime ); } }
        catch(InterruptedException ie){}

        Suggestion sgn = null;
        byte[] snd = null;
        int x = 0;
        int y = 0;
        while( !m_bExit )//&& m_iReadPosition >= 0 )
        {
            //play( 100 );
            //nextScreen( );
            //if ( m_bVoice ) playRow( 0, 100 );
            int dy = 1;
            if ( m_bVoice )
            {
                m_bNextRead = false;
                m_g.setColor( 0, 0, 255 );//blue
                if ( sgn == null )
                {
                    sgn = getSuggestion( m_pStrRows, x, 0 );
                    snd = m_word.getSound( sgn.m_str );
                }
                //highlight suggestion
                getSuggestion( m_pStrRows, x, 0 );
                repaint();
                //play suggestion
                byte[] buf = snd;
                playBuff( buf, 100 );
                //prepare next suggestion
                dy = sgn.m_y - y;
                x = sgn.m_x;
                if ( dy < m_iMaxRows )
                {
                    sgn = getSuggestion( m_pStrRows, x, dy );
                    snd = m_word.getSound( sgn.m_str );
                }
                else
                    m_bVoice = false;
                //wait while play buffer, if need
                if ( !m_bNextRead )
                {
                    try { synchronized( m_waitRead ) { m_waitRead.wait(); } }
                    catch(InterruptedException ie){}
                }

                //stopPlayer();
                
                /*dy = sgn.m_y;
                x = sgn.m_x;*/
            }
            try { synchronized( m_waitScroll ) { m_waitScroll.wait( m_lScrollWaitTime ); } }
            catch(InterruptedException ie){}

            for( y = 0; y < dy; y++ )
            {
                if ( m_bEndBook ) break;
                nextRow();
            }
            //y = dy;
        }
        savePosition();
        try { m_is.close(); } catch (IOException ex) { }
    }
    private void init()
    {
        m_g = getGraphics();
        m_g.setFont( Font.getFont( Font.FACE_PROPORTIONAL, Font.STYLE_PLAIN, Font.SIZE_SMALL ) );
        m_font = m_g.getFont();
        m_iTextWidth = getWidth() - TEXT_LEFT_SHIFT - PROGRESS_BAR_WIDTH;
        m_iMaxRows = getHeight() / m_font.getHeight();
        m_pStrRows = new String[m_iMaxRows];
        m_pPositionRows = new int[m_iMaxRows + 1];
        m_strNextWord = null;
        m_g.drawString( "restore position", getWidth() / 2, getHeight() / 2 - m_font.getHeight(), m_g.HCENTER | m_g.BOTTOM );
        repaint();
        restorePosition();
        //m_iReadPosition = m_iShowPosition;
        m_iNextRowPosition = m_iShowPosition;
        setStreamPosition( m_iNextRowPosition );
        m_g.drawString( "load dictionary", getWidth() / 2, getHeight() / 2, m_g.HCENTER | m_g.BOTTOM );
        repaint();
        m_word = new Word();
        m_dic = new Dictionary();
        m_dic.loadDictionary();
    }
    private void savePosition()
    {
        byte[] data = new byte[4];
        Word.intToByte( m_iShowPosition, data, 0 );
        /*data[0] = (byte)(m_iShowPosition & 0x000000ff);
        data[1] = (byte)((m_iShowPosition>>8) & 0x000000ff);
        data[2] = (byte)((m_iShowPosition>>16) & 0x000000ff);
        data[3] = (byte)((m_iShowPosition>>24) & 0x000000ff);*/
        RecordStore rs = null;
        try{ RecordStore.deleteRecordStore( "Reader" ); }
        catch(RecordStoreException e){}
        try
        {
            rs = RecordStore.openRecordStore( "Reader", true );
            rs.addRecord( data, 0, data.length );
            if (rs != null) rs.closeRecordStore();
        }
        catch(RecordStoreException e){ }
    }
    private void restorePosition()
    {
        RecordStore rs = null;
        m_iShowPosition = 0;
        try
        {
            if( RecordStore.listRecordStores() != null )
            {
                rs = RecordStore.openRecordStore("Reader", false);
                byte data[] = new byte [rs.getRecordSize(rs.getNextRecordID() - 1)];
                data = rs.getRecord(rs.getNextRecordID() - 1);
                m_iShowPosition = ( data[0] & 0x000000ff )
                        | ( ( data[1] << 8 ) & 0x0000ff00 )
                        | ( ( data[2] << 16 ) & 0x00ff0000 )
                        | ( ( data[3] << 24 ) & 0xff000000 );
                if (rs != null) rs.closeRecordStore();
            }
        }
        catch(RecordStoreException e) { }
    }
    private boolean setStreamPosition(int offset)
    {
        try
        {
            m_is = getClass().getResourceAsStream( Reader.BOOK );
            m_iBookLength = m_is.available();
            int d = m_is.read();
            m_encoding = Reader.UNICODE;
            if ( d < 0xff )
            {
                m_encoding = Reader.ANSI;
                m_is.close();
                m_is = getClass().getResourceAsStream( Reader.BOOK );
            }
            else
            {
                //if ( d == 0xef ) m_encoding = Reader.UTF_8;
                d = m_is.read();
                m_iShowPosition >>= 1;
                m_iShowPosition <<= 1;
                offset >>= 1;
                offset <<= 1;
            }
            if ( m_is.skip( offset ) < 0 ) { m_bExit = true; }
        }
        catch(Exception e)
        {
            m_g.drawString( "book not found", getWidth() / 2, getHeight() - m_font.getHeight(), m_g.HCENTER | m_g.BOTTOM );
            repaint();
            m_bExit = true;
        }
        return !m_bExit;
    }
    private String getNextWord() throws IOException, UnsupportedEncodingException
    {
        String str = "";
        do
        {
            char ch = ' ';
            if ( m_encoding == Reader.UNICODE )
            {
                byte[] data = new byte[2];
                if ( m_is.read( data ) <= 0 )
                {
                    m_bEndBook = true;
                    str += Reader.END_BOOK;
                    break;
                }
                ch = (char)( ( data[0] & 0xff ) | ( ( data[1] << 8 ) & 0xff00 ) );
            } else if ( m_encoding == Reader.ANSI )
            {
                int data = m_is.read( );
                if ( data <= 0 )
                {
                    m_bEndBook = true;
                    str += Reader.END_BOOK;
                    break;
                }
                if ( data < 128) ch = (char)data;
                else ch = Reader.WIN1251_TO_UNICODE.charAt( data - 128 );
            } else
            {
                //int data = m_is.read( );
                //if ( data <= 0 )
                {
                    m_bEndBook = true;
                    str += Reader.UNSUPPORTED_ENCODING;
                    break;
                }
            }
            str += String.valueOf( ch );
        }
        while ( !( str.endsWith( " " )
                || str.endsWith( "," )
                || str.endsWith( "." )
                || str.endsWith( ";" )
                || str.endsWith( "!" )
                || str.endsWith( "?" )
                || str.endsWith( "\r" )
                || str.endsWith( "\n" ) ) );
        return str;
    }
    private String prepareRow(int y, String word)
    {
        int x_pos = 0;
        String str = word;
        m_pStrRows[y] = "";
        try
        {
            do
            {
                if ( str != null )
                {
                    while ( str.equals( "\r" ) || str.equals( "\n" ) )
                    { str = getNextWord(); }
                    if ( !str.endsWith( " ") ) str = str.trim();
                    int word_width = m_font.stringWidth( str );
                    if ( x_pos + word_width >= m_iTextWidth ) break;
                    m_pStrRows[y] += str;
                    x_pos += word_width;
                }
                str = getNextWord();
            }
            while( x_pos < m_iTextWidth && !m_bEndBook
                    && !str.equals( "" )
                    && !str.equals( "\r" )
                    && !str.equals( "\n" ) );
        }
        catch (UnsupportedEncodingException e){}
        catch (IOException ie){}
        return str;
    }
    private void prepareText()
    {
        m_pPositionRows[0] = m_iShowPosition;
        for(int y = 0; y < m_iMaxRows; y++)
        {
            m_pStrRows[y] = "";
            if ( !m_bEndBook )
            {
                m_strNextWord = prepareRow( y, m_strNextWord);
                m_iNextRowPosition += m_pStrRows[y].length();
                if ( m_encoding != Reader.ANSI ) { m_iNextRowPosition += m_pStrRows[y].length(); }
                m_pPositionRows[y + 1] = m_iNextRowPosition;
            }
        }
    }
    private void drawText()
    {
        m_g.setColor(255, 255, 255);
        m_g.fillRect( 0, 0, getWidth(), getHeight() );
        //draw progress bar
        m_g.setColor( 0, 255, 0);
        m_g.fillRect( getWidth() - PROGRESS_BAR_WIDTH, 0, getWidth(), getHeight() );
        m_g.setColor( 255, 0, 0);
        m_g.fillRect( getWidth() - PROGRESS_BAR_WIDTH, 0, getWidth(), getHeight() * m_iNextRowPosition / m_iBookLength );
        //
        m_g.setColor(0, 0, 0);
        for(int y = 0; y < m_iMaxRows; y++)
        {
            m_g.drawString( m_pStrRows[y], TEXT_LEFT_SHIFT, y * m_font.getHeight(), Graphics.LEFT | Graphics.TOP );
        }
        repaint();
    }
    private void nextScreen()
    {
        if ( m_bEndBook ) return;
        m_iShowPosition = m_iNextRowPosition;
        prepareText();
        drawText();
    }
    private void nextRow()
    {
        if ( m_bEndBook ) return;
        m_iShowPosition = m_pPositionRows[1];
        for (int y = 1; y < m_iMaxRows; y++)
        {
            m_pStrRows[y - 1] = m_pStrRows[y];
        }
        m_strNextWord = prepareRow( m_iMaxRows - 1, m_strNextWord );
        m_iNextRowPosition += m_pStrRows[m_iMaxRows - 1].length();
        if ( m_encoding != Reader.ANSI ) m_iNextRowPosition += m_pStrRows[m_iMaxRows - 1].length();
        drawText();
    }
    private void stopPlayer()
    {
        try
        {
            if ( m_player != null )
            {
                if( m_player.getState() == Player.STARTED )
                    m_player.stop();
                if( m_player.getState() == Player.PREFETCHED )
                    m_player.deallocate();
                if( m_player.getState() == Player.REALIZED || m_player.getState() == Player.UNREALIZED)
                    m_player.close();
            }
        }
        catch ( MediaException me ){}
        m_player = null;
    }

    public void pausePlayer()
    {
       try
       {
            if( m_player != null && m_player.getState() == Player.STARTED)
                m_player.stop();
            else
                stopPlayer();
        } 
        catch( MediaException me ) { m_player = null; }
    }
    private void playBuff(byte[] snd, int volume)
    {
       try
       {
            InputStream bis = new ByteArrayInputStream( snd );
            m_player = Manager.createPlayer( bis, "audio/X-wav" );
            m_player.addPlayerListener( this );
            m_player.realize();
            setLevel( volume );//max 100
            m_player.prefetch();
            m_player.start();
        }
        catch(MediaException me){}
        catch(IOException ie){}
    }
    /*
    private void playRow(int y, int volume)
    {
       stopPlayer();
       m_g.setColor( 0, 0, 255 );//blue
       int Xpos = 0;
       for ( int x = 0; x < m_pStrRows[y].length(); )
       {
            String word = m_word.getWord(m_pStrRows[y], x);
            m_g.drawString( word, Xpos, y * m_font.getHeight(), Graphics.LEFT | Graphics.TOP );
            repaint();
            x += word.length();
            int wordLength = m_font.stringWidth( word );
            word = m_dic.getPronunciation( word );
            byte[] buf = m_word.getSound( word );
            playBuff( buf, volume );

            try { synchronized( m_waitRead ) { m_waitRead.wait(); } }
            catch(InterruptedException ie){}

            stopPlayer();

            Xpos += wordLength;
       }
    }
    */
    private Suggestion getSuggestion(String[] pStrRows, int x, int y)
    {
        String str = pStrRows[y].substring( 0, x );
        int Xpos = m_font.stringWidth( str ) + TEXT_LEFT_SHIFT;
        str = "";
        for ( ; y < pStrRows.length; y++)
        {
            for( ; x < pStrRows[y].length(); )
            {
                String word = m_word.getWord( pStrRows[y].substring( x ), 0 );
                m_g.drawString( word, Xpos, y * m_font.getHeight(), Graphics.LEFT | Graphics.TOP );
                int dot = word.indexOf( "." );
                if ( dot < 0 ) dot = word.indexOf( ";" );
                if ( dot < 0 ) dot = word.indexOf( "!" );
                if ( dot < 0 ) dot = word.indexOf( "?" );
                if ( dot > 0 )
                {
                    if ( str.lastIndexOf( ' ' ) > 0 )
                    {
                        dot++;
                        Suggestion sgn = new Suggestion();
                        word = word.substring( 0, dot );
                        sgn.m_x = x + word.length();
                        str += m_dic.getPronunciation( word );
                        sgn.m_str = str;
                        sgn.m_y = y;
                        return sgn;
                    }
                }
                Xpos += m_font.stringWidth( word );
                x += word.length();
                str += m_dic.getPronunciation( word );
            }
            x = 0;
            Xpos = TEXT_LEFT_SHIFT;//0;
        }
        Suggestion sgn = new Suggestion();
        sgn.m_str = str;
        sgn.m_x = x;
        sgn.m_y = y;
        return sgn;
    }
    /*
    private Suggestion playSuggestion(String[] pStrRows, int x, int y, int volume)
    {
        m_g.setColor( 0, 0, 255 );//blue
        Suggestion sgn = getSuggestion( m_pStrRows, x, y );
        repaint();

        byte[] buf = m_word.getSound( sgn.m_str );
        playBuff( buf, volume );

        try { synchronized( m_waitRead ) { m_waitRead.wait(); } }
        catch(InterruptedException ie){}

        stopPlayer();
        return sgn;
    }
    */
    private void setLevel(int level)
    {
       VolumeControl vc = (VolumeControl)m_player.getControl( "VolumeControl" );
       if( vc != null ) { vc.setLevel( level ); }
    }
}
