import java.io.*;
/*
 * Word.java
 *
 * Created on 11  2007 ., 23:39
 *
 */

/**
 *
 * @author V*D*V
 */
public class Word
{
    private static final int RIFF = 0;
    private static final int RIFF_SIZE = 4;
    private static final int WAVE_FMT = 8;
    private static final int EXT = 16;
    private static final int FORMAT_TAG = 20; 
    private static final int CHANNELS = 22; 
    private static final int SAMPLES_PER_SEC = 24; 
    private static final int AVG_BYTES_PER_SEC = 28; 
    private static final int BLOCK_ALIGN = 32;
    private static final int BITS_PER_SAMPLE = 34; 
    private static final int SDATA = 36;
    private static final int DATA_SIZE = 40;
    private static final byte[] wav_header_11025_8_mono = {'R', 'I', 'F', 'F',//sRIFF [4]
                            0, 0, 0, 0,//riff_size//full_size - 8
                            'W', 'A', 'V', 'E', 'f', 'm', 't', ' ',//sWAVEfmt[8]
                            0x10, 00, 00, 00,//nExt
                            0x01, 00,//wFormatTag
                            0x01, 00,//nChannels
                            0x11, 0x2B, 00, 00,//nSamplesPerSec
                            0x11, 0x2B, 00, 00,//nAvgBytesPerSec
                            0x01, 00,//nBlockAlign
                            0x08, 00,//wBitsPerSample
                            'd', 'a', 't', 'a',//sdata[4]
                            00, 00, 00, 00};//data_size//full_size - 44
    private static final int MAX_SEARCH_SOUND = 2;//4;
    private static final int MIX_LENGTH = 0 * 11025 / 1000 ;
    //private static final int TRANSLIT_LETTER_LENGTH = 3;
    private static final String[] m_pRusLat =
    {
        "  a", "  b", "  v", "  g", "  d",
        "  e", " yo", " zh", "  z", "  i",
        " iy", "  k", "  l", "  m", "  n",
        "  o", "  p", "  r", "  s", "  t",
        "  u", "  f", "  h", " ts", " ch",
        " sh", " sc", " tz", "  y", " mz",
        " ye", " yu", " ya"
    };
    private static final String[] m_pDigits =
    {
       "", "", "", "", "", "", "", "", "", ""
    };
    private static final String[] m_pTens_10_19 =
    {
       "", "", "", "", "", "", "", "", "", ""
    };
    private static final String[] m_pTens_20_90 =
    {
       "", "", "", "", "", "", "", ""
    };
    private static final String[] m_pHundreds =
    {
       "", "", "", "", "", "", "", "", ""
    };
    private class Vocal
    {// helper class
        public int letters_count = 0;
        public byte[] data = null;
    }

    public Word()
    {
    }
    public String getWord(String str, int offset)
    {
        int next = str.indexOf( " ", offset );
        String word = " ";
        if ( next < 0 ) { word = str.substring( offset ); }
        else if ( next <= offset ) { return word; }
        else word = str.substring( offset, next + 1 );
        return word;
    }
    public byte[] getSound(String str)
    {
        //int sound_size = str.length() * LETTER_MAX_SOUND_SIZE;
        byte[] sound = new byte[wav_header_11025_8_mono.length];
        System.arraycopy( wav_header_11025_8_mono, 0, sound, 0, wav_header_11025_8_mono.length );
        for ( int i = 0; i < str.length(); )
        {
            Vocal voc = getLettersSound( str, i );
            int count = 1;
            if ( voc != null )
            {//do sound mix
                byte[] old_sound = sound;
                int mixLength = Math.min( MIX_LENGTH, voc.data.length );
                int mixPos = old_sound.length - mixLength;
                if ( mixPos > wav_header_11025_8_mono.length )
                {
                    sound = new byte[mixPos + voc.data.length];
                    System.arraycopy( old_sound, 0, sound, 0, old_sound.length );
                    for ( int mix = 0; mixPos < old_sound.length; mixPos++ )
                    {
                        int s1 = ( ( (int)sound[mixPos]) & 0x000000ff );
                        int s2 = ( ( (int)voc.data[mix++]) & 0x000000ff );
                        int sum = s1 + s2 - 128;
                        //int sum = sound[mixPos] + voc.data[mix++] + 128;
                        if ( sum < 0 ) sum = 0;
                        else if ( sum > 255 ) sum = 255;
                        sound[mixPos] = (byte)sum;
                    }
                }
                else
                {
                    sound = new byte[old_sound.length + voc.data.length];
                    System.arraycopy( old_sound, 0, sound, 0, old_sound.length );
                    mixLength = 0;
                }
                System.arraycopy( voc.data, mixLength, sound, old_sound.length, voc.data.length - mixLength);
                count = voc.letters_count;
            }
            i += count;
        }
        int riff_size = sound.length - 8;
        int data_size = sound.length - wav_header_11025_8_mono.length;
        intToByte( riff_size, sound, RIFF_SIZE );
        intToByte( data_size, sound, DATA_SIZE );
        return sound;
    }
    public static void intToByte(int number, byte[] data, int offset)
    {
        data[offset++] = (byte)(number & 0x000000ff);
        data[offset++] = (byte)((number >> 8) & 0x000000ff);
        data[offset++] = (byte)((number >> 16) & 0x000000ff);
        data[offset] = (byte)((number >> 24) & 0x000000ff);
    }
    private Vocal getLettersSound( String str, int offset )
    {
        Vocal voc = new Vocal();
        InputStream is = null;
        if ( str.charAt( offset ) >= '0' && str.charAt( offset ) <= '1' ) return getNumberSound( str, offset );
        int count = Math.min( MAX_SEARCH_SOUND, str.length() - offset );
        for( ; count > 0; count-- )
        {
            String strPh = str.substring( offset, offset + count );
            strPh = translit( strPh );
            is = getClass().getResourceAsStream( strPh + ".W" );
            if ( is != null ) break;
        }
        if ( count <= 0 ) return null;
        try
        {
            int size = is.available();
            //skip header
            //size -= wav_header_11025_8_mono.length;
            //is.skip( wav_header_11025_8_mono.length );
            voc.data = new byte[size];
            voc.letters_count = count;
            is.read( voc.data );
        }
        catch(IOException e) { voc = null; }
        return voc;
    }
    private String translit(String str)
    {
        String translit = "";
        for (int i = 0; i < str.length(); i++)
        {
            char letter = str.charAt( i );
            String s = String.valueOf( letter );
            int indx = -1;
            if ( letter == '' ) indx = 6;
            else if ( letter == '' ) indx = 6;
            else if ( letter >= '' && letter <= '' )
            {
                indx = letter - '';
                if ( letter > '' ) indx++;
            } else if ( letter >= '' && letter <= '' )
            {
                indx = letter - '';
                if ( letter > '' ) indx++;
            }
            if ( indx >= 0 ) s = m_pRusLat[indx];
            translit += s;
        }
        return translit;
    }
    private Vocal getNumberSound(String str, int offset)
    {
        String num = str.substring( offset, str.indexOf( " ", offset ) );
        int k = 0;
        for ( ; k < num.length(); k++ )
        {//check for digits
            if ( num.charAt( k ) > '9' || num.charAt( k ) < '0' ) break;
        }//cut first digits if need
        if ( k < num.length() ) { num = num.substring( 0, k ); }
        //
        byte[] data = null;
        offset = 0;
        int triads = ( num.length() - 1 ) / 3;
        for ( int i = triads * 3; i >= 0; i -= 3 )
        {
            String digit = num.substring( offset, num.length() - i );
            offset += digit.length();
            byte[] olddata = null;
            byte[] bh = null;
            byte[] bt = null;
            byte[] bd = null;
            for( int d = 3 - digit.length(); d < 3; d++ )
            {
                switch( d )
                {
                    case 0:
                        bh = getDigit( digit, "00" );
                        digit = digit.substring( 1 );
                        break;
                    case 1:
                        bt = getDigit( digit, "0");
                        if ( digit.charAt( 0 ) < '2' ) d++;
                        digit = digit.substring( 1 );
                        break;
                    case 2:
                        bd = getDigit( digit, "" );
                        break;
                }
            }
            int size = 0;
            if ( bh != null ) size += bh.length;
            if ( bt != null ) size += bt.length;
            if ( bd != null ) size += bd.length;
            if ( data != null ) olddata = data;
            if ( olddata != null ) size += olddata.length;
            data = new byte[size];
            size = 0;
            if ( olddata != null )
            {
                System.arraycopy( olddata, 0, data, 0, olddata.length );
                size = olddata.length;
            }
            if ( bh != null )
            {
                System.arraycopy( bh, 0, data, size, bh.length );
                size += bh.length;
            }
            if ( bt != null )
            {
                System.arraycopy( bt, 0, data, size, bt.length );
                size += bt.length;
            }
            if ( bd != null )
            {
                System.arraycopy( bd, 0, data, size, bd.length );
            }
        }
        Vocal voc = new Vocal();
        voc.data = data;
        voc.letters_count = num.length();
        return voc;
    }
    private byte[] getDigit(String dig, String shift)
    {//dig "200", shift "00"
        //if ( tens.length() > shift.length() + 1 ) return null;
        if ( dig.length() == 2 && dig.charAt( 0 ) < '2' )
        {
            shift = String.valueOf( dig.charAt( 1 ) );
        }
        dig = String.valueOf( dig.charAt( 0 ) ) + shift;
        InputStream is = getClass().getResourceAsStream( dig + ".W" );
        if ( is == null ) return readDigit( dig, shift );
        byte[] digdata = null;
        try
        {
            digdata = new byte[is.available()/* - wav_header_11025_8_mono.length*/];
            //is.skip( wav_header_11025_8_mono.length );
            is.read( digdata );
        }
        catch (IOException ex) {}
        if ( is != null ) try { is.close(); } catch (IOException ex) {}
        return digdata;
    }
    private byte[] readDigit( String dig, String shift )
    {
        String word = "";
        byte[] digdata = null;
        switch( shift.length() )
        {
            case 0:
                word = m_pDigits[dig.charAt( 0 ) - '0'];
                break;
            case 1:
                if ( dig.charAt( 0 ) < '2' )
                {
                    word = m_pTens_10_19[dig.charAt( 1 ) - '0'];
                    break;
                }
                word = m_pTens_20_90[dig.charAt( 0 ) - '2'];
                break;
            case 2:
                word = m_pHundreds[dig.charAt( 0 ) - '1'];
                break;
        }
        digdata = getSound( word );
        return digdata;
    }
}
